@@ -57,10 +57,18 @@ module Data.PriorityQueue.FingerTree (
 import qualified Data.FingerTree as FT
 import Data.FingerTree (FingerTree, (<|), (|>), (><), ViewL(..), Measured(..))
 
-import Control.Arrow ((***))
+import Prelude hiding (null)
+#if MIN_VERSION_base(4,8,0)
+import qualified Prelude (null)
+#else
 import Data.Foldable (Foldable(foldMap))
 import Data.Monoid
-import Prelude hiding (null)
+#endif
+#if MIN_VERSION_base(4,9,0)
+import Data.Semigroup
+#endif
+import Control.Arrow ((***))
+import Data.List (unfoldr)
 
 data Entry k v = Entry k v
 
@@ -72,13 +80,23 @@ instance Foldable (Entry k) where
 
 data Prio k v = NoPrio | Prio k v
 
+#if MIN_VERSION_base(4,9,0)
+instance Ord k => Semigroup (Prio k v) where
+    (<>) = unionPrio
+#endif
+
 instance Ord k => Monoid (Prio k v) where
-    mempty                  = NoPrio
-    x `mappend` NoPrio      = x
-    NoPrio `mappend` y      = y
-    x@(Prio kx _) `mappend` y@(Prio ky _)
-      | kx <= ky            = x
-      | otherwise           = y
+    mempty  = NoPrio
+#if !(MIN_VERSION_base(4,11,0))
+    mappend = unionPrio
+#endif
+
+unionPrio :: Ord k => Prio k v -> Prio k v -> Prio k v
+x `unionPrio` NoPrio      = x
+NoPrio `unionPrio` y      = y
+x@(Prio kx _) `unionPrio` y@(Prio ky _)
+  | kx <= ky            = x
+  | otherwise           = y
 
 instance Ord k => Measured (Prio k v) (Entry k v) where
     measure (Entry k v) = Prio k v
@@ -89,14 +107,38 @@ newtype PQueue k v = PQueue (FingerTree (Prio k v) (Entry k v))
 instance Ord k => Functor (PQueue k) where
     fmap f (PQueue xs) = PQueue (FT.fmap' (fmap f) xs)
 
+-- | In ascending order of keys.
 instance Ord k => Foldable (PQueue k) where
     foldMap f q = case minView q of
         Nothing -> mempty
         Just (v, q') -> f v `mappend` foldMap f q'
+#if MIN_VERSION_base(4,8,0)
+    null (PQueue q) = FT.null q
+#endif
 
+#if MIN_VERSION_base(4,9,0)
+instance Ord k => Semigroup (PQueue k v) where
+    (<>) = union
+#endif
+
+-- | 'empty' and 'union'
 instance Ord k => Monoid (PQueue k v) where
     mempty = empty
+#if !(MIN_VERSION_base(4,11,0))
     mappend = union
+#endif
+
+instance (Ord k, Eq v) => Eq (PQueue k v) where
+    xs == ys = assocs xs == assocs ys
+
+-- | Lexicographical ordering
+instance (Ord k, Ord v) => Ord (PQueue k v) where
+    compare xs ys = compare (assocs xs) (assocs ys)
+
+-- | In ascending key order
+instance (Ord k, Show k, Show v) => Show (PQueue k v) where
+    showsPrec p xs = showParen (p > 10) $
+        showString "fromList " . shows (assocs xs)
 
 -- | /O(1)/. The empty priority queue.
 empty :: Ord k => PQueue k v
@@ -179,3 +221,7 @@ minViewWithKey (PQueue q)
 below :: Ord k => k -> Prio k v -> Bool
 below _ NoPrio = False
 below k (Prio k' _) = k' <= k
+
+-- | /O(n)/. Key-value pairs in ascending key order.
+assocs :: Ord k => PQueue k v -> [(k, v)]
+assocs = unfoldr minViewWithKey
